from typing import Tuple

from .protocol import Motor as MotorNumber, Parameter, read_command, parse_packet


class Motor:
    def __init__(self, number: MotorNumber):
        self.number = number

    async def get_position(self) -> Tuple[int, int]:
        cmd = read_command(self.number, Parameter.Position)
        _, _, target, feedback = await self.request_read(cmd)
        return target, feedback

    async def get_pwm_status(self) -> Tuple[int, int]:
        cmd = read_command(self.number, Parameter.PwmStatus)
        _, _, target, feedback = await self.request_read(cmd)
        return target, feedback

    async def get_Kp(self) -> int:
        return await self._read_coeff(Parameter.Kp)

    async def get_Ki(self) -> int:
        return await self._read_coeff(Parameter.Ki)

    async def get_Kd(self) -> int:
        return await self._read_coeff(Parameter.Kd)

    async def get_Ks(self) -> int:
        return await self._read_coeff(Parameter.Ks)

    async def _read_coeff(self, k: Parameter) -> int:
        cmd = read_command(self.number, k)
        _, _, kv = await self.request_read(cmd)
        return kv

    async def request_read(self, cmd: bytes) -> Any:
        # Send via serial
        # Receive response
        return parse_packet(cmd)  # Dummy

    async def _send_command(self, cmd: bytes) -> None:
        ...



import asyncio
import serial_asyncio
import logging

from typing import Any, List
from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE

from .loggable import Loggable
from .protocol import Client, Protocol, Motor, Parameter, set_command, DEFAULT_BAUDRATE


class MotorStatus:
    motor: Motor
    target: int
    feedback: int
    pwm: int
    status: int

    def __init__(
        self,
        motor: Motor,
        *,
        target: int = 0,
        feedback: int = 0,
        pwm: int = 0,
        status: int = 0,
    ) -> None:
        self.motor = motor
        self.target = target
        self.feedback = feedback
        self.pwm = pwm
        self.status = status

    def __str__(self) -> str:
        return f"{self.motor.name} target {self.target:3d} feedback {self.feedback:3d} pwm {self.pwm:3d} status {self.status}"


"""
Interface Type - Serial
ComPort = the Arduino ComPort, you can find it in Windows Device Manager.
BitsPerSec - 500000
Data Bits - 8
Parity - None
Stop Bit -1
Output Bit Range - 10
Output Type - Binary
"""


class Box(Loggable):
    """
    Class representing the Simulator Motor Control box
    """

    _loop: asyncio.AbstractEventLoop
    _client: Client
    _motors: List[MotorStatus]

    def __init__(
        self,
        *,
        device: str,
        loop: asyncio.AbstractEventLoop = None,
        baudrate: int = DEFAULT_BAUDRATE,
    ) -> None:
        super().__init__()
        self.set_logger(logging.getLogger("BOX"))

        if not loop:
            loop = asyncio.get_event_loop()

        self._loop = loop
        self._client = Client(
            loop,
            position_cb=self._position_received,
            pwm_status_cb=self._pwm_status_received,
        )
        self._motors = [
            MotorStatus(Motor.A),
            MotorStatus(Motor.B),
            MotorStatus(Motor.C),
        ]
        proto = Protocol(self._client)
        _, _ = loop.run_until_complete(
            serial_asyncio.create_serial_connection(
                loop,
                lambda: proto,
                device,
                baudrate=baudrate,
                bytesize=EIGHTBITS,
                parity=PARITY_NONE,
                stopbits=STOPBITS_ONE,
            )
        )

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    async def get_version_async(self) -> int:
        packet = await self._client.make_read_request(b"[ver]", "v")
        return packet[2]

    def get_version(self) -> int:
        return self._loop.run_until_complete(self.get_version_async())

    async def read_param_async(self, motor: Motor, param: Parameter) -> Any:
        _, _, *args = await self._client.read_parameter(motor, param)
        return args

    def read_param(self, motor: Motor, param: Parameter) -> Any:
        return self._loop.run_until_complete(self.read_param_async(motor, param))

    def delay(self, delay: float) -> None:
        self._loop.run_until_complete(asyncio.sleep(delay))

    def save_settings(self) -> None:
        self._client.send_command(b"[sav]")

    def enable_feedback(self, motor: Motor) -> None:
        self.log_info(f"Enable feedback for {motor.name} {motor.value}")
        self._client.send_command(bytes(f"[mo{motor.value}]", "ascii"))

    def enable_motor(self, motor: Motor) -> None:
        self._client.send_command(bytes(f"[en{motor.value}]", "ascii"))

    def enable_motors(self) -> None:
        self._client.send_command(b"[ena]")

    def disable_feedback(self) -> None:
        self.log_info(f"Disable feedback for motors")
        self._client.send_command(b"[mo0]")

    def set_position(self, motor: Motor, pos: int) -> None:
        self._client.send_command(set_command(motor, Parameter.Position, pos))

    def set_parameter(self, motor: Motor, param: Parameter, *args) -> None:
        self._client.send_command(set_command(motor, param, *args))

    def _position_received(self, motor: Motor, target: int, feedback: int) -> None:
        ms = self._motors[motor.value - 1]
        ms.target = target
        ms.feedback = feedback
        self.log_info(ms)

    def _pwm_status_received(self, motor: Motor, pwm: int, status: int) -> None:
        ms = self._motors[motor.value - 1]
        ms.pwm = pwm
        ms.status = status
        self.log_info(ms)



"""
Packet Sent            Description                                                                                              Response Packet
-------------------------------------------------------------------------------------------------------------------------------------------
[Axx],[Bxx],[Cxx]  Send position updates for Motor 1,2,3 where xx is the binary position limitted to range 0-1024            None
[Dxx],[Exx],[Fxx]  Send the Kp parameter for motor 1,2,3 where xx is the Kp binary value (restrict to 0 - 1000)              None
[Gxx],[Hxx],[Ixx]  Send the Ki parameter for motor 1,2,3 where xx is the Ki binary value (restrict to 0 - 1000)              None
[Jxx],[Kxx],[Lxx]  Send the Kd parameter for motor 1,2,3 where xx is the Kd binary value (restrict to 0 - 1000)              None
[Mxx],[Nxx],[Oxx]  Send the Ks parameter for motor 1,2,3 where xx is the Ks (d term smoothing parameter between 1 and 20)    None
[Pxy],[Qxy],[Rxy]  Send the PWMmin and PWMmax values x is the PWMmin and y is PWMmax each being in range 0-255              None
                    PWMmax should always be greater than or equal to PWMmin
[Sxy],[Txy],[Uxy]  Send the Motor Min/Max Limits (x) and Input Min/Max Limits (y) (Note same value used for Min and Max)    None
[Vxy],[Wxy],[Xxy]  Send the Feedback dead zone (x) and the PWM reverse duty (y) for each motor                              None
[mo1],[mo2],[mo3]  Request continous motor position, feedback, pwm and status data (all packets sent every 15ms)            [Axy][Bxy][Cxy][axy][bxy][cxy]
[mo0]              Disable continuous feedback from previous [mo1],[mo2], or [mo3] command                                  None
[ena]              Enable all motors                                                                                        None
[sav]              Save parameters to non-volatile memory                                                                    None
[ver]              Request the SMC3 software version, returned value is mult by 100.  (ie: 101 is ver 1.01)                  [vyy]
[rdA],[rdB],[rdC]  Request motor target and feedback for Motor 1,2, or 3 parameters scaled 0-255.                            [Axy][Bxy][Cxy]
[rda],[rdb],[rdc]  Request motor pwm and status data for Motor 1,2, or 3 parameters scaled 0-255.                            [Axy][Bxy][Cxy]
[rdD],[rdE],[rdF]  Request the Kp parameter for motor 1,2,3 where xx is the Kp value multiplied by 100                      [Dxx][Exx][Fxx]
[rdG],[rdH],[rdI]  Request the Ki parameter for motor 1,2,3 where xx is the Ki value multiplied by 100                      [Gxx][Hxx][Ixx]
[rdJ],[rdK],[rdL]  Request the Kd parameter for motor 1,2,3 where xx is the Kd value multiplied by 100                      [Jxx][Kxx][Lxx]


Note the Arduino SMC3 code does not check values sent are within valid ranges.  It is the job of the host application to keep values within
sensible limits as recommended above.
"""

import asyncio
import logging
import datetime

from enum import Enum
from typing import Tuple, Any, Dict, Callable

from .loggable import Loggable

PACKET_LEN = 5
BYTE_ORDER = "big"

DEFAULT_BAUDRATE = 500000
DEFAULT_TIMEOUT = datetime.timedelta(seconds=1)


class Motor(Enum):
    A = 1
    B = 2
    C = 3


class Parameter(Enum):
    Unknown = ""
    Position = "A"
    Kp = "D"
    Ki = "G"
    Kd = "J"
    Ks = "M"
    PWMinMax = "P"
    MinMax = "S"
    FBDeadZone = "V"
    PwmStatus = "a"
    Version = "v"

    def __init__(self, v: str):
        self.code = v

    @classmethod
    def _missing_(cls, value):
        p = Parameter.__new__(Parameter, "")
        p.code = value
        return p


READ_ONLY_PARAMS = [Parameter.PwmStatus, Parameter.Version]

COMMAND_ARGC = {
    Parameter.PWMinMax: 2,
    Parameter.MinMax: 2,
    Parameter.FBDeadZone: 2,
}

COMMAND_ARG_LIMITS = {
    Parameter.Position: (0, 1024),
    Parameter.Kp: (0, 1000),
    Parameter.Ki: (0, 1000),
    Parameter.Kd: (0, 1000),
    Parameter.Ks: (1, 20),
}

PARAM_VALUE_COUNT = {
    Parameter.Position: 2,
    Parameter.PwmStatus: 2,
    Parameter.MinMax: 2,
    Parameter.PWMinMax: 2,
    Parameter.FBDeadZone: 2,
}


def param_to_char(motor: Motor, param: Parameter) -> str:
    return chr(ord(param.value) + motor.value - 1)


def char_to_param(char: str) -> Tuple[Motor, Parameter]:
    return byte_to_param(ord(char))


def byte_to_param(byte: int) -> Tuple[Motor, Parameter]:
    start = ord("A")
    if byte >= ord("a"):
        start = ord("a")
    motor = Motor((byte - start) % 3 + 1)
    param = Parameter(chr((byte - start) // 3 * 3 + start))
    return motor, param


def read_command(motor: Motor, param: Parameter) -> bytes:
    return bytes(f"[rd{param_to_char(motor, param)}]", "ascii")


def set_command(motor: Motor, param: Parameter, *args) -> bytes:
    argc = 1
    if param in READ_ONLY_PARAMS:
        raise TypeError(f"Paramenter {param.name} is read-only")
    if param in COMMAND_ARGC:
        argc = COMMAND_ARGC[param]
    if len(args) != argc:
        raise TypeError(
            f"Invalid argument count for parameter {param.name}, got {len(args)} required {argc}"
        )
    limits = param in COMMAND_ARG_LIMITS and COMMAND_ARG_LIMITS[param] or None

    if argc == 2:
        limits = (0, 255)
    if limits:
        for a in args:
            if a < limits[0] or limits[1] < a:
                raise ValueError(
                    f"Value {a} out of range {limits} for param {param.name}"
                )

    return format_value(param_to_char(motor, param), *args)


def format_value(code: str, *args) -> bytes:
    argc = len(args)
    if argc < 1 or 2 < argc:
        raise ValueError(f"Invalid number of values {argc}")
    cmd = bytes(f"[{code}", "ascii")
    if argc == 1:
        cmd = cmd + args[0].to_bytes(2, BYTE_ORDER) + b"]"
    elif argc == 2:
        cmd = cmd + args[0].to_bytes() + args[1].to_bytes() + b"]"
    return cmd


def parse_packet(packet: bytes) -> Tuple[Motor, Parameter]:
    if len(packet) != PACKET_LEN:
        raise ValueError(f"Invalid packet size {len(packet)}")
    motor, param = byte_to_param(packet[1])
    value_count = 1
    if param in PARAM_VALUE_COUNT:
        value_count = PARAM_VALUE_COUNT[param]
    if value_count == 1:
        return (
            chr(packet[1]),
            motor,
            param,
            int.from_bytes(packet[2:4], byteorder=BYTE_ORDER, signed=False),
        )
    elif value_count == 2:
        return chr(packet[1]), motor, param, packet[2], packet[3]
    raise ValueError(f"Invalid value count setup for param {param.name}: {value_count}")


class Client(Loggable):
    _loop: asyncio.AbstractEventLoop
    _transport: asyncio.Transport
    _outstanding: Dict[str, asyncio.Future]

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        position_cb: Callable[[Motor, int, int], None] = None,
        pwm_status_cb: Callable[[Motor, int, int], None] = None,
    ) -> None:
        super().__init__()
        self.set_logger(logging.getLogger("CLIENT"))
        self._loop = loop
        self._transport = None
        self._outstanding = dict[str, asyncio.Future]()
        self._position_cb = position_cb
        self._pwm_status_sb = pwm_status_cb

    def send_command(self, cmd: bytes) -> None:
        self._transport.write(cmd)

    async def wait_for_packet(
        self,
        packet_type: str,
        timeout: datetime.timedelta = DEFAULT_TIMEOUT,
    ) -> Any:
        if packet_type in self._outstanding:
            self.log_warning(f"Already waiting for packet `{packet_type}`")
        fut = self._loop.create_future()
        self._outstanding[packet_type] = fut
        return await asyncio.wait_for(fut, timeout.total_seconds())

    async def read_parameter(
        self,
        motor: Motor,
        param: Parameter,
        timeout: datetime.timedelta = DEFAULT_TIMEOUT,
    ) -> Any:
        return await self.make_read_request(
            read_command(motor, param), param_to_char(motor, param), timeout=timeout
        )

    async def make_read_request(
        self,
        cmd: bytes,
        wait_for: str,
        timeout: datetime.timedelta = DEFAULT_TIMEOUT,
    ) -> Any:
        self.send_command(cmd)
        return await self.wait_for_packet(wait_for, timeout=timeout)

    def packet_recieved(
        self, packet_type: str, motor: Motor, param: Parameter, *args
    ) -> None:
        self.log_debug(f"Received packet '{packet_type}' {motor} {param} {args}")
        if packet_type in self._outstanding:
            self._outstanding[packet_type].set_result((motor, param, *args))
            del self._outstanding[packet_type]
        # Check for position and status callback
        elif param == Parameter.Position:
            if self._position_cb:
                self._position_cb(motor, args[0], args[1])
        elif param == Parameter.PwmStatus:
            if self._pwm_status_sb:
                self._pwm_status_sb(motor, args[0], args[1])


class Protocol(asyncio.Protocol, Loggable):
    def __init__(self, client: Client):
        super().__init__()
        self.set_logger(logging.getLogger("PROTO"))

        self._transport = None
        self._client = client
        self._buffer = bytes()

    def connection_made(self, transport) -> None:
        self._transport = transport
        self._client._transport = transport
        self.log_debug(f"port opened {self._transport}")

    def data_received(self, data) -> None:
        self.log_debug(f"data received {repr(data)}")
        if not data:
            return
        self._buffer = self._buffer + data
        self.process_buffer()

    def process_buffer(self) -> None:
        if len(self._buffer) < 5:
            return

        buffer = self._buffer[:5]
        packet = parse_packet(buffer)
        self._client.packet_recieved(*packet)

        self._buffer = self._buffer[5:]
        self.process_buffer()

    def connection_lost(self, exc) -> None:
        self.log_debug("port closed")
        self._transport.loop.stop()

    def pause_writing(self) -> None:
        self.log_debug("pause writing")
        self.log_debug(self._transport.get_write_buffer_size())

    def resume_writing(self) -> None:
        self.log_debug(self._transport.get_write_buffer_size())
        self.log_debug("resume writing")



import logging


class Loggable:
    def set_logger(self, logger: logging.Logger) -> None:
        self._logger = logger
        self.log_debug = self._logger.debug
        self.log_info = self._logger.info
        self.log_warning = self._logger.warning
        self.log_error = self._logger.error



from distutils.core import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="smc3",  # How you named your package folder (MyLib)
    packages=["smc3"],  # Chose the same as "name"
    version="0.1.1",  # Start with a small number and increase it with every change you make
    license="Apache-2.0",  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description="Simulation Motor Controller interface",  # Give a short description about your library
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Sergei A. Fedorov",  # Type in your name
    author_email="sergei.a.fedorov@gmail.com",  # Type in your E-Mail
    url="https://github.com/zmij/pysmc3",  # Provide either the link to your github or to your website
    download_url="https://github.com/zmij/pysmc3/archive/refs/tags/v0.1.tar.gz",  # I explain this later on
    keywords=[
        "motion simulation",
        "motor control",
        "diagnostic tools",
    ],
    install_requires=[  # I get to this in a second
        "pyserial-asyncio",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Intended Audience :: Developers",
        "Topic :: Games/Entertainment :: Simulation",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)
